@@ -6,6 +6,7 @@ package pwru
 import (
 	"errors"
 	"fmt"
+	"log/slog"
 
 	"github.com/cilium/ebpf"
 	"golang.org/x/sys/unix"
@@ -20,6 +21,7 @@ func listBpfProgs(typ ebpf.ProgramType) ([]*ebpf.Program, error) {
 	)
 
 	var progs []*ebpf.Program
+	var progNames []string
 	for id, err = ebpf.ProgramGetNextID(id); err == nil; id, err = ebpf.ProgramGetNextID(id) {
 		prog, err := ebpf.NewProgramFromID(id)
 		if err != nil {
@@ -28,15 +30,21 @@ func listBpfProgs(typ ebpf.ProgramType) ([]*ebpf.Program, error) {
 
 		if prog.Type() == typ {
 			progs = append(progs, prog)
+			progNames = append(progNames, prog.String())
 		} else {
 			_ = prog.Close()
 		}
 	}
+	slog.Debug("Found bpf progs", "names", progNames)
 
 	if !errors.Is(err, unix.ENOENT) { // Surely err != nil
 		return nil, err
 	}
 
+	if len(progs) == 0 {
+		slog.Info("No bpf progs found with type", "type", typ.String())
+	}
+
 	return progs, nil
 }
 
