@@ -7,7 +7,7 @@ import (
 	"context"
 	"errors"
 	"fmt"
-	"log"
+	"log/slog"
 	"os"
 	"slices"
 	"strings"
@@ -16,6 +16,7 @@ import (
 
 	"github.com/cheggaaa/pb/v3"
 	"github.com/cilium/ebpf"
+	"github.com/cilium/ebpf/btf"
 	"github.com/cilium/ebpf/link"
 	"golang.org/x/sync/errgroup"
 )
@@ -64,9 +65,9 @@ func attachKprobes(ctx context.Context, bar *pb.ProgressBar, kprobes []Kprobe) (
 }
 
 // AttachKprobes attaches kprobes concurrently.
-func AttachKprobes(ctx context.Context, bar *pb.ProgressBar, kps []Kprobe, batch uint) (links []link.Link, ignored int) {
+func AttachKprobes(ctx context.Context, bar *pb.ProgressBar, kps []Kprobe, batch uint) (links []link.Link, ignored int, err error) {
 	if batch == 0 {
-		log.Fatal("--filter-kprobe-batch must be greater than 0")
+		return nil, 0, fmt.Errorf("--filter-kprobe-batch must be greater than 0")
 	}
 
 	var kprobes []Kprobe
@@ -117,7 +118,7 @@ func AttachKprobes(ctx context.Context, bar *pb.ProgressBar, kps []Kprobe, batch
 	}
 
 	if err := errg.Wait(); err != nil {
-		log.Fatalf("Attaching kprobes: %v\n", err)
+		return nil, 0, fmt.Errorf("attaching kprobes: %v", err)
 	}
 
 	return
@@ -125,7 +126,7 @@ func AttachKprobes(ctx context.Context, bar *pb.ProgressBar, kps []Kprobe, batch
 
 // DetachKprobes detaches kprobes concurrently.
 func (k *kprober) DetachKprobes() {
-	log.Println("Detaching kprobes...")
+	slog.Info("Detaching kprobes...")
 
 	links := k.links
 	bar := pb.StartNew(len(links))
@@ -162,7 +163,7 @@ func (k *kprober) DetachKprobes() {
 }
 
 // AttachKprobeMulti attaches kprobe-multi serially.
-func AttachKprobeMulti(ctx context.Context, bar *pb.ProgressBar, kprobes []Kprobe, a2n Addr2Name) (links []link.Link, ignored int) {
+func AttachKprobeMulti(ctx context.Context, bar *pb.ProgressBar, kprobes []Kprobe, a2n Addr2Name) (links []link.Link, ignored int, err error) {
 	links = make([]link.Link, 0, len(kprobes))
 
 	for _, kp := range kprobes {
@@ -188,10 +189,10 @@ func AttachKprobeMulti(ctx context.Context, bar *pb.ProgressBar, kprobes []Kprob
 		}
 
 		opts := link.KprobeMultiOptions{Addresses: addrs}
-		l, err := link.KprobeMulti(kp.Prog, opts)
+		l, err0 := link.KprobeMulti(kp.Prog, opts)
 		bar.Add(len(kp.HookFuncs))
-		if err != nil {
-			log.Fatalf("Opening kprobe-multi: %s\n", err)
+		if err0 != nil {
+			return nil, 0, fmt.Errorf("opening kprobe-multi: %s", err0)
 		}
 
 		links = append(links, l)
@@ -200,13 +201,13 @@ func AttachKprobeMulti(ctx context.Context, bar *pb.ProgressBar, kprobes []Kprob
 	return
 }
 
-func NewKprober(ctx context.Context, funcs Funcs, coll *ebpf.Collection, a2n Addr2Name, useKprobeMulti bool, batch uint) *kprober {
+func NewKprober(ctx context.Context, funcs Funcs, coll *ebpf.Collection, a2n Addr2Name, useKprobeMulti bool, batch uint) (*kprober, error) {
 	msg, probeMethod := "kprobe", "kprobe"
 	if useKprobeMulti {
 		msg = "kprobe-multi"
 		probeMethod = "kprobe_multi"
 	}
-	log.Printf("Attaching kprobes (via %s)...\n", msg)
+	slog.Info("Attaching kprobes", "via", msg)
 
 	ignored := 0
 	bar := pb.StartNew(len(funcs))
@@ -228,26 +229,32 @@ func NewKprober(ctx context.Context, funcs Funcs, coll *ebpf.Collection, a2n Add
 	k.kprobeBatch = batch
 
 	if !useKprobeMulti {
-		l, i := AttachKprobes(ctx, bar, pwruKprobes, batch)
+		l, i, err := AttachKprobes(ctx, bar, pwruKprobes, batch)
+		if err != nil {
+			return nil, err
+		}
 		k.links = l
 		ignored += i
 	} else {
-		l, i := AttachKprobeMulti(ctx, bar, pwruKprobes, a2n)
+		l, i, err := AttachKprobeMulti(ctx, bar, pwruKprobes, a2n)
+		if err != nil {
+			return nil, err
+		}
 		k.links = l
 		ignored += i
 	}
 	bar.Finish()
 	select {
 	case <-ctx.Done():
-		return &k
+		return &k, nil
 	default:
 	}
-	log.Printf("Attached (ignored %d)\n", ignored)
+	slog.Info("Attached", "ignored", ignored)
 
-	return &k
+	return &k, nil
 }
 
-func NewNonSkbFuncsKprober(nonSkbFuncs []string, funcs Funcs, coll *ebpf.Collection) *kprober {
+func NewNonSkbFuncsKprober(nonSkbFuncs []string, funcs Funcs, bpfmapFuncs map[string]*btf.FuncProto, coll *ebpf.Collection) *kprober {
 	slices.Sort(nonSkbFuncs)
 	nonSkbFuncs = slices.Compact(nonSkbFuncs)
 
@@ -264,12 +271,56 @@ func NewNonSkbFuncsKprober(nonSkbFuncs []string, funcs Funcs, coll *ebpf.Collect
 			continue
 		}
 
+		if _, ok := bpfmapFuncs[fn]; ok {
+			if strings.HasSuffix(fn, "_lookup_elem") {
+				kp, err := link.Kprobe(fn, coll.Programs["kprobe_bpf_map_lookup_elem"], nil)
+				if err != nil {
+					if !errors.Is(err, os.ErrNotExist) {
+						slog.Warn("Failed to attach bpf_map_lookup_elem kprobe", "func", fn, "error", err)
+					}
+					continue
+				}
+				k.links = append(k.links, kp)
+
+				krp, err := link.Kretprobe(fn, coll.Programs["kretprobe_bpf_map_lookup_elem"], nil)
+				if err != nil {
+					if errors.Is(err, os.ErrNotExist) {
+						slog.Warn("Failed to open bpf_map_lookup_elem kretprobe", "func", fn, "error", err)
+					}
+					continue
+				}
+				k.links = append(k.links, krp)
+
+			} else if strings.HasSuffix(fn, "_update_elem") {
+				kp, err := link.Kprobe(fn, coll.Programs["kprobe_bpf_map_update_elem"], nil)
+				if err != nil {
+					if errors.Is(err, os.ErrNotExist) {
+						slog.Warn("Failed to open bpf_map_update_elem kprobe", "func", fn, "error", err)
+					}
+					continue
+				}
+				k.links = append(k.links, kp)
+
+			} else if strings.HasSuffix(fn, "_delete_elem") {
+				kp, err := link.Kprobe(fn, coll.Programs["kprobe_bpf_map_delete_elem"], nil)
+				if err != nil {
+					if errors.Is(err, os.ErrNotExist) {
+						slog.Warn("Failed to open bpf_map_delete_elem kprobe", "func", fn, "error", err)
+					}
+					continue
+				}
+				k.links = append(k.links, kp)
+			}
+
+			continue
+		}
+
 		kp, err := link.Kprobe(fn, coll.Programs["kprobe_skb_by_stackid"], nil)
 		if err != nil {
 			if errors.Is(err, os.ErrNotExist) {
 				continue
 			}
-			log.Printf("Opening non-skb-kprobe %s: %s\n", fn, err)
+			slog.Warn("Opening non-skb-kprobe", "func", fn, "error", err)
 		} else {
 			k.links = append(k.links, kp)
 		}
