@@ -5,10 +5,12 @@
 package pwru
 
 import (
+	"bytes"
+	"encoding/binary"
 	"encoding/json"
 	"errors"
 	"fmt"
-	"log"
+	"log/slog"
 	"net"
 	"os"
 	"path/filepath"
@@ -43,6 +45,7 @@ type output struct {
 	printSkbMap    *ebpf.Map
 	printShinfoMap *ebpf.Map
 	printStackMap  *ebpf.Map
+	printBpfmapMap *ebpf.Map
 	addr2name      Addr2Name
 	skbMetadata    []*SkbMetadata
 	xdpMetadata    []*SkbMetadata
@@ -50,28 +53,29 @@ type output struct {
 	kprobeMulti    bool
 	kfreeReasons   map[uint64]string
 	ifaceCache     map[uint64]map[uint32]string
+	procCache      map[int]string
 }
 
 // outputStructured is a struct to hold the data for the json output
 type jsonPrinter struct {
-	Skb         string      `json:"skb,omitempty"`
-	Shinfo      string      `json:"skb_shared_info,omitempty"`
-	Cpu         uint32      `json:"cpu,omitempty"`
-	Process     string      `json:"process,omitempty"`
-	Func        string      `json:"func,omitempty"`
-	CallerFunc  string      `json:"caller_func,omitempty"`
-	Time        interface{} `json:"time,omitempty"`
-	Netns       uint32      `json:"netns,omitempty"`
-	Mark        uint32      `json:"mark,omitempty"`
-	Iface       string      `json:"iface,omitempty"`
-	Proto       uint16      `json:"proto,omitempty"`
-	Mtu         uint32      `json:"mtu,omitempty"`
-	Len         uint32      `json:"len,omitempty"`
-	Cb          [5]uint32   `json:"cb,omitempty"`
-	Tuple       *jsonTuple  `json:"tuple,omitempty"`
-	TunnelTuple *jsonTuple  `json:"tunnel_tuple,omitempty"`
-	Stack       interface{} `json:"stack,omitempty"`
-	SkbMetadata interface{} `json:"skb_metadata,omitempty"`
+	Skb         string     `json:"skb,omitempty"`
+	Shinfo      string     `json:"skb_shared_info,omitempty"`
+	Cpu         uint32     `json:"cpu,omitempty"`
+	Process     string     `json:"process,omitempty"`
+	Func        string     `json:"func,omitempty"`
+	CallerFunc  string     `json:"caller_func,omitempty"`
+	Time        any        `json:"time,omitempty"`
+	Netns       uint32     `json:"netns,omitempty"`
+	Mark        uint32     `json:"mark,omitempty"`
+	Iface       string     `json:"iface,omitempty"`
+	Proto       uint16     `json:"proto,omitempty"`
+	Mtu         uint32     `json:"mtu,omitempty"`
+	Len         uint32     `json:"len,omitempty"`
+	Cb          [5]uint32  `json:"cb,omitempty"`
+	Tuple       *jsonTuple `json:"tuple,omitempty"`
+	TunnelTuple *jsonTuple `json:"tunnel_tuple,omitempty"`
+	Stack       any        `json:"stack,omitempty"`
+	SkbMetadata any        `json:"skb_metadata,omitempty"`
 }
 
 type jsonTuple struct {
@@ -92,7 +96,7 @@ func centerAlignString(s string, width int) string {
 	return fmt.Sprintf("%s%s%s", strings.Repeat(" ", leftPadding), s, strings.Repeat(" ", rightPadding))
 }
 
-func NewOutput(flags *Flags, printSkbMap, printShinfoMap, printStackMap *ebpf.Map, addr2Name Addr2Name, skbMds, xdpMds []*SkbMetadata, kprobeMulti bool, btfSpec *btf.Spec) (*output, error) {
+func NewOutput(flags *Flags, printSkbMap, printShinfoMap, printStackMap, printBpfmapMap *ebpf.Map, addr2Name Addr2Name, skbMds, xdpMds []*SkbMetadata, kprobeMulti bool, btfSpec *btf.Spec) (*output, error) {
 	writer := os.Stdout
 
 	if flags.OutputFile != "" {
@@ -105,14 +109,14 @@ func NewOutput(flags *Flags, printSkbMap, printShinfoMap, printStackMap *ebpf.Ma
 
 	reasons, err := getKFreeSKBReasons(btfSpec)
 	if err != nil {
-		log.Printf("Unable to load packet drop reasons: %v", err)
+		slog.Warn("Unable to load packet drop reasons", "error", err)
 	}
 
 	var ifs map[uint64]map[uint32]string
 	if flags.OutputMeta {
 		ifs, err = getIfaces()
 		if err != nil {
-			log.Printf("Failed to retrieve all ifaces from all network namespaces: %v. Some iface names might be not shown.", err)
+			slog.Warn("Failed to retrieve all ifaces from all network namespaces. Some iface names might be not shown.", "error", err)
 		}
 	}
 
@@ -122,6 +126,7 @@ func NewOutput(flags *Flags, printSkbMap, printShinfoMap, printStackMap *ebpf.Ma
 		printSkbMap:    printSkbMap,
 		printShinfoMap: printShinfoMap,
 		printStackMap:  printStackMap,
+		printBpfmapMap: printBpfmapMap,
 		addr2name:      addr2Name,
 		skbMetadata:    skbMds,
 		xdpMetadata:    xdpMds,
@@ -129,6 +134,7 @@ func NewOutput(flags *Flags, printSkbMap, printShinfoMap, printStackMap *ebpf.Ma
 		kprobeMulti:    kprobeMulti,
 		kfreeReasons:   reasons,
 		ifaceCache:     ifs,
+		procCache:      map[int]string{},
 	}, nil
 }
 
@@ -167,14 +173,14 @@ func (o *output) PrintHeader() {
 }
 
 // PrintJson prints the event in JSON format
-func (o *output) PrintJson(event *Event) {
+func (o *output) PrintJson(event *Event) error {
 	// crate an instance of the outputStructured struct to hold the data
 	d := &jsonPrinter{}
 
 	// add the data to the struct
 	d.Skb = fmt.Sprintf("%#x", event.SkbAddr)
 	d.Cpu = event.CPU
-	d.Process = getExecName(int(event.PID))
+	d.Process = o.getExecName(int(event.PID))
 	d.Func = getOutFuncName(o, event, event.Addr)
 	if o.flags.OutputCaller {
 		d.CallerFunc = o.addr2name.findNearestSym(event.CallerAddr)
@@ -245,8 +251,9 @@ func (o *output) PrintJson(event *Event) {
 
 	err := encoder.Encode(d)
 	if err != nil {
-		log.Fatalf("Error encoding JSON: %s", err)
+		return fmt.Errorf("error encoding JSON: %s", err)
 	}
+	return nil
 }
 
 func getAbsoluteTs() string {
@@ -263,7 +270,11 @@ func getRelativeTs(event *Event, o *output) uint64 {
 	return ts
 }
 
-func getExecName(pid int) string {
+func (o *output) getExecName(pid int) string {
+	if name, ok := o.procCache[pid]; ok {
+		return name
+	}
+
 	p, err := ps.FindProcess(pid)
 	execName := fmt.Sprintf("<empty>:%d", pid)
 	if err == nil && p != nil {
@@ -275,6 +286,8 @@ func getExecName(pid int) string {
 			execName = string(bexecName)
 		}
 	}
+
+	o.procCache[pid] = execName
 	return execName
 }
 
@@ -422,11 +435,14 @@ func fprintWithPadding(writer *os.File, data string, maxLenSeen *int) {
 }
 
 func (o *output) Print(event *Event) {
+	var sb strings.Builder
+	sb.Grow(256)
+
 	if o.flags.OutputTS == "absolute" {
-		fmt.Fprintf(o.writer, "%-12s ", getAbsoluteTs())
+		sb.WriteString(fmt.Sprintf("%-12s ", getAbsoluteTs()))
 	}
 
-	execName := getExecName(int(event.PID))
+	execName := o.getExecName(int(event.PID))
 
 	ts := event.Timestamp
 	if o.flags.OutputTS == "relative" {
@@ -435,54 +451,75 @@ func (o *output) Print(event *Event) {
 
 	outFuncName := getOutFuncName(o, event, event.Addr)
 
-	fmt.Fprintf(o.writer, "%-18s %-3s %-16s", fmt.Sprintf("%#x", event.SkbAddr),
-		fmt.Sprintf("%d", event.CPU), fmt.Sprintf("%s", execName))
+	sb.WriteString(fmt.Sprintf("%-18s %-3s %-16s", fmt.Sprintf("%#x", event.SkbAddr),
+		fmt.Sprintf("%d", event.CPU), execName))
 	if o.flags.OutputTS != "none" {
-		fmt.Fprintf(o.writer, " %-16d", ts)
+		sb.WriteString(fmt.Sprintf(" %-16d", ts))
 	}
 	o.lastSeenSkb[event.SkbAddr] = event.Timestamp
 
 	if o.flags.OutputMeta {
-		fmt.Fprintf(o.writer, " %s", getMetaData(event, o))
+		sb.WriteString(" ")
+		sb.WriteString(getMetaData(event, o))
 		if o.flags.FilterTraceTc || o.flags.OutputSkbCB {
-			fmt.Fprintf(o.writer, " %s", getCb(event))
+			sb.WriteString(" ")
+			sb.WriteString(getCb(event))
 		}
 	}
 
 	if o.flags.OutputTuple {
-		fprintWithPadding(o.writer, getTupleData(event, o.flags.OutputTCPFlags), &maxTupleLengthSeen)
+		tupleData := getTupleData(event, o.flags.OutputTCPFlags)
+		if len(tupleData) > maxTupleLengthSeen {
+			maxTupleLengthSeen = len(tupleData)
+		}
+		sb.WriteString(fmt.Sprintf(" %-*s", maxTupleLengthSeen, tupleData))
 	}
 
 	if o.flags.OutputCaller {
-		fprintWithPadding(o.writer, outFuncName, &maxFuncLengthSeen)
-		fmt.Fprintf(o.writer, " %s", o.addr2name.findNearestSym(event.CallerAddr))
+		if len(outFuncName) > maxFuncLengthSeen {
+			maxFuncLengthSeen = len(outFuncName)
+		}
+		sb.WriteString(fmt.Sprintf(" %-*s", maxFuncLengthSeen, outFuncName))
+		sb.WriteString(" ")
+		sb.WriteString(o.addr2name.findNearestSym(event.CallerAddr))
 	} else {
-		fmt.Fprintf(o.writer, " %s", outFuncName)
+		sb.WriteString(" ")
+		sb.WriteString(outFuncName)
 	}
 
 	if event.Type != eventTypeTracingXdp && len(o.skbMetadata) > 0 {
-		outputSkbMetadata(o.writer, o.skbMetadata, event.SkbMetadata[:])
+		var metaBuf strings.Builder
+		outputSkbMetadata(&metaBuf, o.skbMetadata, event.SkbMetadata[:])
+		sb.WriteString(metaBuf.String())
 	} else if event.Type == eventTypeTracingXdp && len(o.xdpMetadata) > 0 {
-		outputSkbMetadata(o.writer, o.xdpMetadata, event.SkbMetadata[:])
+		var metaBuf strings.Builder
+		outputSkbMetadata(&metaBuf, o.xdpMetadata, event.SkbMetadata[:])
+		sb.WriteString(metaBuf.String())
 	}
 
 	if o.flags.OutputStack && event.PrintStackId > 0 {
-		fmt.Fprintf(o.writer, "%s", getStackData(event, o))
+		sb.WriteString(getStackData(event, o))
 	}
 
 	if o.flags.OutputSkb {
-		fmt.Fprintf(o.writer, "%s", getSkbData(event, o))
+		sb.WriteString(getSkbData(event, o))
 	}
 
 	if o.flags.OutputShinfo {
-		fmt.Fprintf(o.writer, "%s", getShinfoData(event, o))
+		sb.WriteString(getShinfoData(event, o))
 	}
 
 	if o.flags.OutputTunnel {
-		fmt.Fprintf(o.writer, " %s", getTuple(event.TunnelTuple, o.flags.OutputTCPFlags))
+		sb.WriteString(" ")
+		sb.WriteString(getTuple(event.TunnelTuple, o.flags.OutputTCPFlags))
 	}
 
-	fmt.Fprintln(o.writer)
+	if o.flags.OutputBpfmap && event.PrintBpfmapId > 0 {
+		sb.WriteString(getBpfMapData(event, o))
+	}
+
+	sb.WriteString("\n")
+	o.writer.Write([]byte(sb.String()))
 }
 
 func (o *output) getIfaceName(netnsInode, ifindex uint32) string {
@@ -641,3 +678,17 @@ func getIfacesInNetNs(path string) (map[uint32]string, error) {
 
 	return ifaces, nil
 }
+
+func getBpfMapData(event *Event, o *output) (bpfMapData string) {
+	id := uint64(event.PrintBpfmapId)
+	b, err := o.printBpfmapMap.LookupBytes(&id)
+	if err != nil {
+		return ""
+	}
+	bpfmap := printBpfmapValue{}
+	if err = binary.Read(bytes.NewBuffer(b), byteorder.Native, &bpfmap); err != nil {
+		return ""
+	}
+	defer o.printBpfmapMap.Delete(&id)
+	return "\n" + bpfmap.String()
+}
