@@ -8,10 +8,11 @@ import (
 	"bufio"
 	"fmt"
 	"iter"
-	"log"
+	"log/slog"
 	"os"
 	"path/filepath"
 	"regexp"
+	"strings"
 
 	"github.com/cilium/ebpf"
 	"github.com/cilium/ebpf/asm"
@@ -25,9 +26,15 @@ type Funcs map[string]int
 // to attach kprobes.
 func getAvailableFilterFunctions() (map[string]struct{}, error) {
 	availableFuncs := make(map[string]struct{})
-	f, err := os.Open("/sys/kernel/debug/tracing/available_filter_functions")
+	var f *os.File
+    var err error
+	// Linux v6.17+ enforce this new path for the tracing functionality
+	f, err = os.Open("/sys/kernel/tracing/available_filter_functions")
 	if err != nil {
-		return nil, fmt.Errorf("failed to open: %v", err)
+		f, err = os.Open("/sys/kernel/debug/tracing/available_filter_functions")
+		if err != nil {
+			return nil, fmt.Errorf("failed to open: %v", err)
+		}
 	}
 	defer f.Close()
 
@@ -43,8 +50,9 @@ func getAvailableFilterFunctions() (map[string]struct{}, error) {
 	return availableFuncs, nil
 }
 
-func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti bool) (Funcs, error) {
+func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti, bpfmap bool) (Funcs, map[string]*btf.FuncProto, error) {
 	funcs := Funcs{}
+	bpfmapFuncs := map[string]*btf.FuncProto{}
 
 	type iterator struct {
 		kmod string
@@ -53,13 +61,13 @@ func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti bool)
 
 	reg, err := regexp.Compile(pattern)
 	if err != nil {
-		return nil, fmt.Errorf("failed to compile regular expression %v", err)
+		return nil, nil, fmt.Errorf("failed to compile regular expression %v", err)
 	}
 
 	var availableFuncs map[string]struct{}
 	availableFuncs, err = getAvailableFilterFunctions()
 	if err != nil {
-		log.Printf("Failed to retrieve available ftrace functions (is /sys/kernel/debug/tracing mounted?): %s", err)
+		slog.Warn("Failed to retrieve available ftrace functions (is /sys/kernel/tracing or /sys/kernel/debug/tracing mounted?)", "error", err)
 	}
 
 	iters := []iterator{{"", spec.All()}}
@@ -67,13 +75,13 @@ func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti bool)
 		path := filepath.Join("/sys/kernel/btf", module)
 		f, err := os.Open(path)
 		if err != nil {
-			return nil, fmt.Errorf("failed to open %s: %v", path, err)
+			return nil, nil, fmt.Errorf("failed to open %s: %v", path, err)
 		}
 		defer f.Close()
 
 		modSpec, err := btf.LoadSplitSpecFromReader(f, spec)
 		if err != nil {
-			return nil, fmt.Errorf("failed to load %s btf: %v", module, err)
+			return nil, nil, fmt.Errorf("failed to load %s btf: %v", module, err)
 		}
 		iters = append(iters, iterator{module, modSpec.All()})
 	}
@@ -81,7 +89,7 @@ func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti bool)
 	for _, it := range iters {
 		for typ, err := range it.iter {
 			if err != nil {
-				return nil, fmt.Errorf("failed to iterate through btf types: %v", err)
+				return nil, nil, fmt.Errorf("failed to iterate through btf types: %v", err)
 			}
 
 			fn, ok := typ.(*btf.Func)
@@ -106,6 +114,17 @@ func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti bool)
 			fnProto := fn.Type.(*btf.FuncProto)
 			i := 1
 			for _, p := range fnProto.Params {
+				if bpfmap && i == 1 {
+					if ptr, ok := p.Type.(*btf.Pointer); ok {
+						if strct, ok := ptr.Target.(*btf.Struct); ok {
+							if strct.Name == "bpf_map" {
+								bpfmapFuncs[fnName] = fnProto
+							}
+						}
+					} else if strings.HasSuffix(p.Name, "map") && p.Type.TypeName() == "u64" {
+						bpfmapFuncs[fnName] = fnProto
+					}
+				}
 				if ptr, ok := p.Type.(*btf.Pointer); ok {
 					if strct, ok := ptr.Target.(*btf.Struct); ok {
 						if strct.Name == "sk_buff" && i <= 5 {
@@ -123,7 +142,7 @@ func GetFuncs(pattern string, spec *btf.Spec, kmods []string, kprobeMulti bool)
 		}
 	}
 
-	return funcs, nil
+	return funcs, bpfmapFuncs, nil
 }
 
 func GetFuncsByPos(funcs Funcs) map[int][]string {
